@@ -30,6 +30,8 @@
 #define SSH_BUFSIZ 131072
 #endif
 
+#define MIN(a,b) (a) < (b) ? (a) : (b)
+
 /* from libssh examples */
 static int verify_knownhost(ssh_session session)
 {
@@ -228,12 +230,14 @@ int ssh_open_url(url_t* urlp)
     args_t *args = args_create();
     args_init(args, 0, NULL);
     args_push_back(args, gvSSHOptions);
-    if (ssh_options_getopt(ftp->session, &args->argc, args->argv) != SSH_OK) {
+    int argc = 0;
+    if (ssh_options_getopt(ftp->session, &argc, args->argv) != SSH_OK) {
       ftp_err(_("Failed to load SSH options from yafcrc config (ssh_options = '%s')\n"), gvSSHOptions);
       ssh_free(ftp->session);
       ftp->session = NULL;
       return -1;
     }
+    args->argc = argc;
     args_destroy(args);
   }
 
@@ -346,7 +350,9 @@ char *ssh_getcurdir(void)
 
 int ssh_chdir(const char *path)
 {
-  char *p = ftp_path_absolute(path);
+  char* tmp = ftp_path_absolute(path);
+  char *p = sftp_canonicalize_path(ftp->sftp_session, tmp);
+  free(tmp);
   bool isdir = false;
 
   /* First check if this file is cached and is a directory, else we
@@ -509,7 +515,14 @@ rdirectory *ssh_read_directory(const char *path)
     if (attrib->group)
       rf->group = xstrdup(attrib->group);
 
-    asprintf(&rf->path, "%s/%s", p, attrib->name);
+    if (asprintf(&rf->path, "%s/%s", p, attrib->name) == -1)
+    {
+      ftp_err(_("Failed to allocate memory.\n"));
+      sftp_closedir(dir);
+      free(p);
+      rdir_destroy(rdir);
+      rfile_destroy(rf);
+    }
     rf->mtime = attrib->mtime;
     rf->date = time_to_string(rf->mtime);
     rf->size = attrib->size;
@@ -588,9 +601,97 @@ void ssh_pwd(void)
   printf("%s\n", ftp->curdir);
 }
 
+static int do_scp_read(ssh_scp scp, const char* infile, FILE* fp,
+                       getmode_t mode, ftp_transfer_func hookf)
+{
+  time_t then = time(NULL) - 1;
+  ftp_set_close_handler();
+
+  if (hookf)
+    hookf(&ftp->ti);
+  ftp->ti.begin = false;
+
+  int rc = ssh_scp_pull_request(scp);
+  if (rc != SSH_SCP_REQUEST_NEWFILE)
+  {
+    ftp_err(_("Failed to start scp download: %s\n"),
+            ssh_get_error(ftp->session));
+    ssh_scp_close(scp);
+    ssh_scp_free(scp);
+    return -1;
+  }
+
+  size_t size = ssh_scp_request_get_size(scp);
+  ssh_scp_accept_request(scp);
+
+  /* read file */
+  char buffer[SSH_BUFSIZ];
+  int r = 0;
+  while (size && (r = ssh_scp_read(scp, buffer, MIN(SSH_BUFSIZ, size))) != SSH_ERROR)
+  {
+    if (ftp_sigints() > 0)
+    {
+      ftp_trace("break due to sigint\n");
+      break;
+    }
+
+    errno = 0;
+    if (fwrite(buffer, r, 1, fp) != 1)
+    {
+      ftp_err(_("Error while writing to file: %s\n"), strerror(errno));
+      ssh_scp_close(scp);
+      ssh_scp_free(scp);
+      return -1;
+    }
+
+    ftp->ti.size += r;
+    if (hookf)
+    {
+      time_t now = time(NULL);
+      if (now > then)
+      {
+        hookf(&ftp->ti);
+        then = now;
+      }
+    }
+    size -= r;
+  }
+
+  if (r == SSH_ERROR)
+  {
+    ftp_err(_("Error while reading from file: %s\n"), ssh_get_error(ftp->session));
+    r = -1;
+  }
+  else
+  {
+    r = ssh_scp_pull_request(scp);
+    if (r != SSH_SCP_REQUEST_EOF)
+      ftp_err(_("Unexpected request: %s %lu\n"), ssh_get_error(ftp->session), size);
+    else
+      r = 0;
+  }
+
+  ssh_scp_close(scp);
+  ssh_scp_free(scp);
+  return r;
+}
+
 static int do_read(const char* infile, FILE* fp, getmode_t mode,
                    ftp_transfer_func hookf, uint64_t offset)
 {
+  if (gvSSHTrySCP && !offset)
+  {
+    /* try to set up a scp connection */
+    ssh_scp scp = ssh_scp_new(ftp->session, SSH_SCP_READ, infile);
+    if (scp != NULL)
+    {
+      int rc = ssh_scp_init(scp);
+      if (rc == SSH_OK)
+        return do_scp_read(scp, infile, fp, mode, hookf);
+      ssh_scp_free(scp);
+    }
+  }
+
   time_t then = time(NULL) - 1;
   ftp_set_close_handler();
 
@@ -689,9 +790,96 @@ int ssh_do_receive(const char *infile, FILE *fp, getmode_t mode,
   return (r == 0 && !ftp->ti.ioerror && !ftp->ti.interrupted) ? 0 : -1;
 }
 
+static int do_scp_write(ssh_scp scp, const char* path, FILE* fp,
+                        ftp_transfer_func hookf)
+{
+  time_t then = time(NULL) - 1;
+  ftp_set_close_handler();
+
+  if (hookf)
+    hookf(&ftp->ti);
+  ftp->ti.begin = false;
+
+  struct stat sb;
+  errno = 0;
+  if (fstat(fileno(fp), &sb) == -1)
+  {
+    ftp_err(_("Couldn't fstat local file: %s\n"), strerror(errno));
+    ssh_scp_free(scp);
+    return -1;
+  }
+
+  int rc = ssh_scp_push_file(scp, path, sb.st_size, sb.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO));
+  if (rc != SSH_OK)
+  {
+    ftp_err(_("Failed to start scp upload: %s\n"),
+            ssh_get_error(ftp->session));
+    ssh_scp_close(scp);
+    ssh_scp_free(scp);
+    return -1;
+  }
+
+  /* read file */
+  char buffer[SSH_BUFSIZ];
+  ssize_t nbytes = 0;
+  errno = 0;
+  while ((nbytes = fread(buffer, sizeof(char), sizeof(buffer), fp)) > 0)
+  {
+    if (ftp_sigints() > 0)
+    {
+      ftp_trace("break due to sigint\n");
+      break;
+    }
+
+    rc = ssh_scp_write(scp, buffer, nbytes);
+    if (rc != SSH_OK)
+    {
+      ftp_err(_("Error while writing to file: %s\n"), ssh_get_error(ftp->session));
+      ssh_scp_close(scp);
+      ssh_scp_free(scp);
+      return -1;
+    }
+
+    ftp->ti.size += nbytes;
+    if (hookf)
+    {
+      time_t now = time(NULL);
+      if (now > then)
+      {
+        hookf(&ftp->ti);
+        then = now;
+      }
+    }
+    errno = 0;
+  }
+
+  if (ferror(fp))
+  {
+    ftp_err(_("Failed to read from file: %s\n"), strerror(errno));
+    rc = -1;
+  }
+
+  ssh_scp_close(scp);
+  ssh_scp_free(scp);
+  return rc;
+}
+
 static int do_write(const char* path, FILE* fp, ftp_transfer_func hookf,
                     uint64_t offset)
 {
+  /* try to set up a scp connection */
+  if (gvSSHTrySCP && !offset)
+  {
+    ssh_scp scp = ssh_scp_new(ftp->session, SSH_SCP_WRITE, path);
+    if (scp != NULL)
+    {
+      int rc = ssh_scp_init(scp);
+      if (rc == SSH_OK)
+        return do_scp_write(scp, path, fp, hookf);
+      ssh_scp_free(scp);
+    }
+  }
+
   time_t then = time(NULL) - 1;
   ftp_set_close_handler();
 
